import os
from dataclasses import dataclass
from typing import Dict, List, Set, Tuple, Union, Optional

import numpy as np
import rasterio

from vpt_core.image.filter_factory import create_filter_by_sequence
from vpt_core.io.vzgfs import get_rasterio_environment, rasterio_open
from vpt_core.segmentation.segmentation_task import SegTask


@dataclass(frozen=True)
class ImageInfo:
    channel: str
    z_layer: int
    full_path: Union[str, os.PathLike]


class ImageSet(Dict[str, Dict[int, np.ndarray]]):
    def z_levels(self) -> Set[int]:
        return set().union(*self.values())

    def as_list(self, key: str) -> List[np.ndarray]:
        return list(self.get(key, {}).values())

    def as_stack(self, order: Optional[List[str]] = None):
        if not order:
            return np.array([np.stack([z_stack[z] for z_stack in self.values()], axis=-1) for z in self.z_levels()])
        return np.array([np.stack([self[k][z] for k in order], axis=-1) for z in self.z_levels()])


def read_tile(window: Tuple[int, int, int, int], path: str, num_tries: int = 5) -> np.ndarray:
    num_retries = num_tries - 1

    for try_number in range(1, num_tries + 1):
        try:
            with get_rasterio_environment(path):
                with rasterio_open(path) as file:
                    image = file.read(
                        1,
                        window=rasterio.windows.Window(window[0], window[1], window[2], window[3]),
                    )
        except OSError:
            if try_number + 1 <= num_tries:
                print(f"Failed to read {path} at {window}. Retrying {try_number}/{num_retries}.")
            continue
        break
    else:
        raise IOError(f"Failed to read {path} at {window}")

    return np.squeeze(image)


def get_segmentation_images(images_info: List[ImageInfo], window_info: Tuple[int, int, int, int]) -> ImageSet:
    images = ImageSet()
    for image_info in images_info:
        if not images.get(image_info.channel):
            images[image_info.channel] = {}
        images[image_info.channel][image_info.z_layer] = read_tile(window_info, str(image_info.full_path))
    return images


def get_prepared_images(task: SegTask, images: ImageSet) -> Tuple[ImageSet, Tuple[float, float]]:
    cur_images = ImageSet()
    scale = None
    for input_info in task.task_input_data:
        image_filter = create_filter_by_sequence(input_info.image_preprocessing)
        task_z = sorted([z for z in images[input_info.image_channel] if z in task.z_layers])
        task_images = [images[input_info.image_channel][z] for z in task_z]
        if len(task_z) == 0:
            raise ValueError(
                f'Invalid experiment: no "{input_info.image_channel}" input images found for the task {task.task_id}'
            )
        filtered_images = image_filter(task_images)
        cur_scale = [
            (im.shape[0] / cur_im.shape[0], im.shape[1] / cur_im.shape[1])
            for im, cur_im in zip(task_images, filtered_images)
        ]
        if min(cur_scale) != max(cur_scale) or (scale is not None and scale != cur_scale[0]):
            raise ValueError(
                "Invalid preprocessing scale: input images for segmentation after postprocessing "
                "should have same sizes"
            )
        scale = cur_scale[0]
        cur_images[input_info.image_channel] = {z: filtered_images[i] for i, z in enumerate(task_z)}
    if scale is None:
        raise ValueError(f"Invalid experiment: no input images found for the task {task.task_id}")

    return cur_images, scale



from __future__ import annotations

from dataclasses import dataclass
from typing import Set, Dict, Callable, Optional

from shapely.geometry.base import BaseGeometry

import numpy as np
from geopandas import GeoDataFrame
from vpt_core.segmentation.geometry_utils import convert_to_multipoly, get_valid_geometry

from vpt_core.segmentation import ResultFields


@dataclass
class SegmentationItem(ResultFields):
    def __init__(self, entity_type: str, df: GeoDataFrame):
        self.entity_type = entity_type
        self.df = df

    def get_entity_id(self) -> int:
        return int(self.df[self.cell_id_field].iloc[0])

    def get_entity_type(self) -> str:
        return self.entity_type

    def set_entity_type(self, entity_type: str) -> None:
        self.entity_type = entity_type
        self.df = self.df.assign(**{self.entity_name_field: entity_type})

    def set_parent(self, parent_entity_type: Optional[str], parent_entity_id: Optional[int]) -> None:
        self.df = self.df.assign(
            **{self.parent_id_field: parent_entity_id, self.parent_entity_field: parent_entity_type}
        )
        self.df.loc[:, self.parent_id_field] = self.df.loc[:, self.parent_id_field].astype("object")

    def get_volume(self):
        return np.sum([x.area for x in self.df[self.geometry_field]])

    def get_z_planes(self) -> Set[int]:
        return set(self.df[self.z_index_field])

    def as_geometries_by_z(self) -> Dict:
        return dict(
            zip(
                self.df[self.z_index_field],
                self.df[self.geometry_field],
            )
        )

    def as_copy(self) -> SegmentationItem:
        return SegmentationItem(self.entity_type, self.df.copy())

    def get_overlapping_volume(self, item: SegmentationItem):
        """
        Calulates intersection area for 3D shapes stored in a geo data frame.
        """
        left_entity = self.as_geometries_by_z()
        right_entity = item.as_geometries_by_z()
        intersection_area = 0
        for z in left_entity:
            if z in right_entity:
                intersection_area += (
                    get_valid_geometry(left_entity[z]).intersection(get_valid_geometry(right_entity[z])).area
                )
        return intersection_area

    @staticmethod
    def process_same_plane_geometries(
        target: GeoDataFrame, arg: GeoDataFrame, op: Callable[[BaseGeometry, BaseGeometry], BaseGeometry]
    ) -> None:
        z_planes_in_both = set(target[SegmentationItem.z_index_field]) & set(arg[SegmentationItem.z_index_field])

        def get_geometry(df: GeoDataFrame, z: int):
            idx = df.loc[df[SegmentationItem.z_index_field] == z].index[0]
            item = df.at[idx, SegmentationItem.geometry_field]
            return idx, get_valid_geometry(item)

        for z in z_planes_in_both:
            # Trims the larger geometry with a small buffer
            ind, valid_item = get_geometry(target, z)

            _, valid_arg = get_geometry(arg, z)

            result_raw = op(valid_item, valid_arg)
            result_geometry = convert_to_multipoly(get_valid_geometry(result_raw))

            if not result_geometry.is_empty:
                # Overwrites large geometry with trimmed geometry
                target.at[ind, SegmentationItem.geometry_field] = result_geometry
            else:
                target.drop(ind, inplace=True)


def intersection(a: SegmentationItem, b: SegmentationItem) -> SegmentationItem:
    result = a.as_copy()
    z_field = SegmentationItem.z_index_field
    result.df.drop(result.df[~result.df[z_field].isin(b.df[z_field])].index, inplace=True)
    SegmentationItem.process_same_plane_geometries(result.df, b.df, lambda a, b: a.intersection(b))
    return result


def difference(a: SegmentationItem, b: SegmentationItem, min_distance: float = 1) -> SegmentationItem:
    result = a.as_copy()
    SegmentationItem.process_same_plane_geometries(
        result.df, b.df, lambda a, b: a.difference(b.buffer(max(min_distance, 1e-10)))
    )
    return result



import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List

from shapely.errors import ShapelyDeprecationWarning

import vpt_core.log as log
from vpt_core.segmentation.seg_result import SegmentationResult


@dataclass
class PolygonParams:
    min_final_area: int = 0
    min_distance_between_entities: int = 2


@dataclass
class SegFusion:
    entity_fusion_strategy: str
    fused_polygon_postprocessing_parameters: PolygonParams = PolygonParams()

    def __post_init__(self):
        if isinstance(self.fused_polygon_postprocessing_parameters, dict):
            self.fused_polygon_postprocessing_parameters = PolygonParams(**self.fused_polygon_postprocessing_parameters)


def run_harmonization(segmentations: List[SegmentationResult], min_distance: int, min_area: int):
    segmentation: SegmentationResult = SegmentationResult.combine_segmentations(segmentations)
    segmentation.make_non_overlapping_polys(min_distance, min_area)
    return segmentation


def run_union_fusion(segmentations: List[SegmentationResult], min_distance: int, min_area: int):
    segmentation: SegmentationResult = SegmentationResult.combine_segmentations(segmentations)
    segmentation.union_intersections(min_distance, min_area)
    return segmentation


def run_larger_fusion(segmentations: List[SegmentationResult], min_distance: int, min_area: int):
    segmentation: SegmentationResult = SegmentationResult.combine_segmentations(segmentations)
    segmentation.larger_resolve_intersections(min_distance, min_area)
    return segmentation


class FusionCallbacks(Enum):
    HARMONIZE = ("harmonize", run_harmonization)
    LARGER = ("larger", run_larger_fusion)
    UNION = ("union", run_union_fusion)


def fuse_task_polygons(
    segmentation_results: List[SegmentationResult],
    fusion_parameters: Dict[str, SegFusion],
):
    log.info("fuse_task_polygons")
    warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
    results = []
    tasks_entities = [res.entity_type for res in segmentation_results]

    for entity_type in set(tasks_entities):
        parameters = fusion_parameters[entity_type].fused_polygon_postprocessing_parameters
        strategy_key = fusion_parameters[entity_type].entity_fusion_strategy.upper()
        if strategy_key not in FusionCallbacks.__members__:
            raise Exception("Invalid fusion strategy")

        cur_results = [seg_res for i, seg_res in enumerate(segmentation_results) if tasks_entities[i] == entity_type]
        seg_result = FusionCallbacks[strategy_key].value[1](
            cur_results,
            parameters.min_distance_between_entities,
            parameters.min_final_area,
        )
        seg_result.set_entity_type(entity_type)
        results.append(seg_result)

    return results



import json
import os
import warnings
from enum import Enum
from typing import Tuple, Callable

import gcsfs
import rasterio.session
import tenacity
from fsspec import AbstractFileSystem, FSTimeoutError
from fsspec.implementations.local import LocalFileSystem
from gcsfs import GCSFileSystem
from rasterio.errors import NotGeoreferencedWarning
from s3fs import S3FileSystem
from tenacity import stop_after_delay, stop_after_attempt, retry_if_exception_type, wait_fixed

from vpt_core import (
    AWS_ACCESS_KEY_VAR,
    AWS_PROFILE_NAME_VAR,
    AWS_SECRET_KEY_VAR,
    GCS_SERVICE_ACCOUNT_KEY_VAR,
    log,
)


class Protocol(Enum):
    LOCAL = 0
    S3 = 1
    GCS = 2


AWS_PROFILE_NAME = None
AWS_ACCESS_KEY = None
AWS_SECRET_KEY = None

GCS_SERVICE_ACCOUNT_KEY = None


def initialize_filesystem(
    aws_access_key=None,
    aws_secret_key=None,
    aws_profile_name=None,
    gcs_service_account_key=None,
):
    global AWS_PROFILE_NAME, AWS_ACCESS_KEY, AWS_SECRET_KEY, GCS_SERVICE_ACCOUNT_KEY

    if all([aws_profile_name, aws_access_key, aws_secret_key]):
        log.warning("Both AWS profile name and access key/secret key pair are specified.")

    AWS_PROFILE_NAME = aws_profile_name if aws_profile_name else os.environ.get(AWS_PROFILE_NAME_VAR)
    AWS_ACCESS_KEY = aws_access_key if aws_access_key else os.environ.get(AWS_ACCESS_KEY_VAR)
    AWS_SECRET_KEY = aws_secret_key if aws_secret_key else os.environ.get(AWS_SECRET_KEY_VAR)

    if gcs_service_account_key:
        GCS_SERVICE_ACCOUNT_KEY = gcs_service_account_key
    else:
        GCS_SERVICE_ACCOUNT_KEY = os.environ.get(GCS_SERVICE_ACCOUNT_KEY_VAR)
        try:
            GCS_SERVICE_ACCOUNT_KEY = json.loads(GCS_SERVICE_ACCOUNT_KEY)
        except Exception:
            pass


def filesystem_for_protocol(protocol: Protocol) -> AbstractFileSystem:
    global AWS_PROFILE_NAME, AWS_ACCESS_KEY, AWS_SECRET_KEY, GCS_SERVICE_ACCOUNT_KEY

    if protocol == Protocol.LOCAL:
        return LocalFileSystem()
    elif protocol == Protocol.S3:
        return S3FileSystem(key=AWS_ACCESS_KEY, secret=AWS_SECRET_KEY, profile=AWS_PROFILE_NAME)
    elif protocol == Protocol.GCS:
        return GCSFileSystem(token=GCS_SERVICE_ACCOUNT_KEY)

    raise NotImplementedError(f"Protocol {protocol} is not supported")


def prefix_for_protocol(protocol: Protocol) -> str:
    if protocol == Protocol.LOCAL:
        return ""
    elif protocol == Protocol.S3:
        return "s3://"
    elif protocol == Protocol.GCS:
        return "gcs://"
    raise NotImplementedError(f"Protocol {protocol} is not supported")


def protocol_path_split(uri: str) -> Tuple[Protocol, str]:
    if uri.startswith("s3://"):
        split = Protocol.S3, uri[5:]
    elif uri.startswith("gcs://"):
        split = Protocol.GCS, uri[6:]
    elif uri.startswith("gs://"):
        split = Protocol.GCS, uri[5:]
    else:
        split = Protocol.LOCAL, uri
    check_access_to_uri(filesystem_for_protocol(split[0]), split[1])
    return split


def check_access_to_uri(fs: AbstractFileSystem, path: str):
    try:
        fs.exists(path)
    except gcsfs.retry.HttpError as e:
        raise ValueError(
            f'{e.message}. Pass credentials or run the command "gcloud auth application-default login" '
            f"to authenticate"
        )


def filesystem_path_split(uri: str) -> Tuple[AbstractFileSystem, str]:
    protocol, path = protocol_path_split(uri)
    fs = filesystem_for_protocol(protocol)

    assert fs is not None

    check_access_to_uri(fs, path)

    return fs, path


def retrying_attempts():
    return tenacity.Retrying(
        stop=(stop_after_delay(60) | stop_after_attempt(5)),
        retry=retry_if_exception_type(FSTimeoutError),
        wait=wait_fixed(10),
    )


def vzg_open(uri: str, mode: str):
    protocol, path = protocol_path_split(uri)
    fs = filesystem_for_protocol(protocol)

    assert fs is not None

    return fs.open(path, mode)


def io_with_retries(uri: str, mode: str, callback: Callable):
    for attempt in retrying_attempts():
        with attempt, vzg_open(uri, mode) as f:
            return callback(f)


def __get_rasterio_session(uri: str):
    warnings.filterwarnings("ignore", category=NotGeoreferencedWarning)
    global AWS_PROFILE_NAME, AWS_ACCESS_KEY, AWS_SECRET_KEY, GCS_SERVICE_ACCOUNT_KEY

    protocol, _ = protocol_path_split(uri)

    if protocol == Protocol.LOCAL:
        return rasterio.session.DummySession()
    elif protocol == Protocol.S3:
        return rasterio.session.AWSSession(
            aws_access_key_id=AWS_ACCESS_KEY,
            aws_secret_access_key=AWS_SECRET_KEY,
            profile_name=AWS_PROFILE_NAME,
        )
    elif protocol == Protocol.GCS:
        return rasterio.session.GSSession(google_application_credentials=GCS_SERVICE_ACCOUNT_KEY)

    raise NotImplementedError(f"Protocol {protocol} is not supported")


def rasterio_open(uri: str):
    protocol, path = protocol_path_split(uri)

    prefix = "/vsigs/" if protocol == Protocol.GCS else prefix_for_protocol(protocol)
    rasterio_path = prefix + path

    return rasterio.open(rasterio_path)


def get_rasterio_environment(uri: str, gdal_cache_size: int = 512000000) -> rasterio.Env:
    return rasterio.Env(__get_rasterio_session(uri), GDAL_CACHEMAX=gdal_cache_size)



from dataclasses import dataclass
from typing import Dict, List


@dataclass(frozen=True)
class SegTask:
    task_id: int
    segmentation_family: str
    entity_types_detected: List[str]
    z_layers: List[int]
    task_input_data: List
    segmentation_parameters: Dict
    segmentation_properties: Dict
    polygon_parameters: Dict



import os
from pathlib import Path

import dotenv

AWS_PROFILE_NAME_VAR = "VPT_AWS_PROFILE"
AWS_ACCESS_KEY_VAR = "VPT_AWS_ACCESS_KEY_ID"
AWS_SECRET_KEY_VAR = "VPT_AWS_SECRET_ACCESS_KEY"
GCS_SERVICE_ACCOUNT_KEY_VAR = "VPT_GCS_SERVICE_ACCOUNT_KEY"

envPath = os.path.join(Path.home(), ".vptenv")

if os.path.exists(envPath):
    dotenv.load_dotenv(envPath)



from vpt_core.io.regex_tools import parse_images_str
from vpt_core.io.vzgfs import vzg_open, retrying_attempts


def _copy_between_filesystems(input_path: str, output_path: str):
    for attempt in retrying_attempts():
        with attempt:
            with vzg_open(input_path, "rb") as i, vzg_open(output_path, "wb") as o:
                data = i.read()
                o.write(data)


def _copy_regex_images(input_regex_path, output_dir, output_separator):
    images = parse_images_str(input_regex_path)
    images = images.images
    if len(images) == 0:
        raise FileNotFoundError(f"Images are not found by regex {input_regex_path}")
    for im in images:
        stain, z, img_path = im.channel, im.z_layer, im.full_path
        _copy_between_filesystems(img_path, output_separator.join([output_dir, f"mosaic_{stain}_z{z}.tif"]))



import logging
import pathlib
import sys
from typing import List

from tqdm import tqdm

_t_id = "/"


class VPTFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        record.tid = _t_id
        record.exc_text, record.exc_info, record.stack_info = None, None, None
        return super().format(record)


vpt_logger = logging.getLogger("vpt")
logger: logging.Logger = vpt_logger
h_stdout = logging.StreamHandler(sys.stdout)
formatter = VPTFormatter("%(asctime)s - %(tid)s - %(levelname)s - %(message)s")
h_stdout.setFormatter(formatter)
file_handlers: List[logging.FileHandler] = []


def set_logger(user_logger: logging.Logger):
    global logger, vpt_logger
    vpt_logger = logger
    logger = user_logger
    set_verbose(True)


def release_logger():
    global logger, vpt_logger, file_handlers
    for fh in file_handlers:
        logger.removeHandler(fh)
        fh.close()
    file_handlers = []
    logger = vpt_logger


def set_process_name(proc_name: str):
    global _t_id
    _t_id = proc_name


def initialize_logger(fname: str, lvl: int = 1, verbose: bool = False) -> None:
    logger.setLevel(lvl * 10)
    if fname:
        pathlib.Path(fname).parent.mkdir(parents=True, exist_ok=True)
        set_log_file(fname)
    if lvl > 1:
        # todo: find a better place for this dask-related calls:
        dask_logger = logging.getLogger("distributed.worker")
        dask_logger.setLevel(logging.ERROR)
    set_verbose(verbose)


def set_log_file(fname: str) -> None:
    fh = logging.FileHandler(filename=fname)
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    file_handlers.append(fh)


def set_verbose(v: bool = True) -> None:
    std_in = is_verbose()
    if std_in == v:
        return
    if v:
        logger.addHandler(h_stdout)
    else:
        logger.removeHandler(h_stdout)


def is_verbose() -> bool:
    return h_stdout in logger.handlers


def show_progress(iterable, *args, **kwargs):
    return iterable if not is_verbose() else tqdm(iterable, *args, **kwargs)


def debug(msg, *args, **kwargs):
    logger.debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    logger.info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    logger.warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    logger.error(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    logger.critical(msg, *args, **kwargs)


def log_system_info():
    import psutil
    from psutil._common import bytes2human

    stat = psutil.virtual_memory()

    text = [f"{name}: {bytes2human(getattr(stat, name))}" for name in ["total", "available", "used", "free"]]
    debug(f"memory stat:{', '.join(text)}")



from dataclasses import dataclass
from typing import Callable, List, Union

import numpy as np

Filter = Callable[[Union[np.ndarray, List[np.ndarray]]], Union[np.ndarray, List[np.ndarray]]]


@dataclass
class Header:
    name: str
    parameters: dict

    def __init__(self, name: str, parameters=None):
        if parameters is None:
            parameters = {}
        self.name = name
        self.parameters = parameters



class ResultFields:
    detection_id_field: str = "ID"
    cell_id_field: str = "EntityID"
    z_index_field: str = "ZIndex"
    geometry_field: str = "Geometry"
    entity_name_field: str = "Type"
    parent_id_field: str = "ParentID"
    parent_entity_field: str = "ParentType"



import re
from dataclasses import dataclass
from typing import Dict, List, Set

from vpt_core import log
from vpt_core.io.vzgfs import (
    filesystem_for_protocol,
    get_rasterio_environment,
    prefix_for_protocol,
    protocol_path_split,
    rasterio_open,
)

default_stain_re = r"(?P<stain>[\w|-]+)"
default_z_re = r"(?P<z>[0-9]+)"
default_images_re = f"mosaic_{default_stain_re}_z{default_z_re}.tif"

pattern_to_re: Dict[str, str] = {"stain": default_stain_re, "z": default_z_re}


@dataclass(frozen=True)
class ImagePath:
    channel: str
    z_layer: int
    full_path: str


@dataclass(frozen=True)
class RegexInfo:
    image_width: int
    image_height: int
    images: Set[ImagePath]


def get_paths_by_regex(regex: str) -> List[str]:
    protocol, regex = protocol_path_split(regex)
    fs = filesystem_for_protocol(protocol)
    parts = regex.split(fs.sep)

    for i in range(len(parts)):
        if "?" in parts[i]:
            break

    root = fs.sep.join(parts[:i])
    parts = parts[i:]

    def walk_req(rootdir, path_regex):
        if len(path_regex) == 0:
            return [rootdir]
        res = []

        try:
            _, dirs, files = next(fs.walk(rootdir, maxdepth=1))
        except StopIteration:
            return res

        if len(path_regex) == 1:
            for filename in files:
                if re.match(path_regex[0], filename):
                    res.append(filename)
            return res

        for dirname in dirs:
            if re.match(path_regex[0], dirname):
                walk_res = walk_req(fs.sep.join([rootdir, dirname]), path_regex[1:])
                for p in walk_res:
                    res.append(fs.sep.join([dirname, p]))

        return res

    paths = [fs.sep.join([root, path]) if root else path for path in walk_req(root, parts)]
    return [path for path in paths if fs.isfile(path)]


def default_for_dir(path: str) -> str:
    fs = filesystem_for_protocol(protocol_path_split(path)[0])
    if fs.isdir(path):
        if path.endswith(fs.sep):
            return path + default_images_re
        return path + fs.sep + default_images_re
    return path


def unify_path_re_str(path: str) -> str:
    path = path.replace(r"\\", "/")
    inside = False

    def scan(a: str) -> str:
        nonlocal inside
        if not inside and a == "(":
            inside = True
        elif inside and a == ")":
            inside = False
        return "/" if not inside and a == "\\" else a

    return "".join([scan(x) for x in path])


def parse_images_str(images_str: str) -> RegexInfo:
    width, height = 0, 0
    images = set()
    images_str = unify_path_re_str(images_str)
    # apply patterns
    images_str = images_str.format(**pattern_to_re)
    # fix directory
    images_str = default_for_dir(images_str)
    paths = get_paths_by_regex(images_str)
    protocol, regex = protocol_path_split(images_str)
    fs = filesystem_for_protocol(protocol)

    if "?P<stain>" not in regex or "?P<z>" not in regex:
        raise ValueError('Bad regular expression: named group "z" or "stain" is missed')

    for path in paths:
        match = re.match(regex, path)
        if match is None:
            log.warning(f"Extracted path {path} does not match the input regular expression")
            continue

        try:
            z = int(match.group("z"))
            stain = match.group("stain")
        except IndexError:
            log.warning(
                f'Regular expression should contain groups "z" and "stain". Path {path} has not that groups '
                f"and will be skipped."
            )
            continue

        full_path = prefix_for_protocol(protocol) + fs.info(path)["name"]

        with get_rasterio_environment(full_path):
            with rasterio_open(full_path) as tif:
                im_height, im_width = tif.height, tif.width
                if width and height and (im_width != width or im_height != height):
                    raise ValueError("Images sizes are not equal")
                width = im_width
                height = im_height

        images.add(ImagePath(stain, z, full_path))

    return RegexInfo(width, height, images)



import json
import os
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from datetime import datetime

import geopandas as gpd
import numpy as np
import pandas as pd
import pyarrow
import pyarrow.parquet as pq
from geopandas.io.arrow import _geopandas_to_arrow

from vpt_core.io.vzgfs import filesystem_path_split
from vpt_core.segmentation.seg_result import SegmentationResult

MIN_ROW_GROUP_SIZE = 1000


@contextmanager
def suppress_output():
    with open(os.devnull, "w") as devnull:
        with redirect_stdout(devnull), redirect_stderr(devnull):
            yield


def make_parent_dirs(path):
    fs, path = filesystem_path_split(path)
    parent = fs.sep.join(path.split(fs.sep)[:-1])
    fs.mkdirs(parent, exist_ok=True)


def save_segmentation_results(gdf: gpd.GeoDataFrame, path: str, max_row_group_size=None):
    if max_row_group_size is None:
        save_geodataframe(gdf, path)
    else:
        id_field = SegmentationResult.cell_id_field
        gdf = gdf.sort_values(id_field)
        row_group_sizes = derive_row_group_sizes(gdf[id_field], max_row_group_size)
        save_geodataframe_with_row_groups(gdf, path, row_group_sizes)


def save_geodataframe(gdf: gpd.GeoDataFrame, path: str) -> None:
    fs, path_inside_fs = filesystem_path_split(path)
    ext = path.split(fs.sep)[-1].split(".")[-1]

    if ext == "geojson":
        with fs.open(path_inside_fs, "wb") as f:
            if gdf.empty:
                f.write(json.dumps({"type": "FeatureCollection", "features": []}).encode("utf-8"))
            else:
                gdf.to_file(f, driver="GeoJSON")
    elif ext == "parquet":
        with fs.open(path_inside_fs, "wb") as f:
            gdf.to_parquet(f)
    else:
        raise ValueError(f"Unknown file extension: {ext}")


def derive_row_group_sizes(entity_id: pd.Series, max_size: int) -> list[int]:
    assert entity_id.is_monotonic_increasing
    vc = entity_id.value_counts(sort=False)

    group_sizes = []
    curr_group_size = 0

    for i in range(len(vc)):
        if curr_group_size + vc.iloc[i] > max_size:
            group_sizes.append(curr_group_size)
            curr_group_size = vc.iloc[i]
        else:
            curr_group_size += vc.iloc[i]

    if curr_group_size > 0:
        group_sizes.append(curr_group_size)

    return group_sizes


def save_geodataframe_with_row_groups(gdf: gpd.GeoDataFrame, path: str, row_group_sizes: list[int]) -> None:
    fs, path_inside_fs = filesystem_path_split(path)
    ext = path.split(fs.sep)[-1].split(".")[-1]

    assert ext == "parquet", "Row groups is a .parquet format specific feature"

    df = gdf.to_wkb()
    schema = _geopandas_to_arrow(gdf).schema

    with fs.open(path_inside_fs, "wb") as f:
        writer = pq.ParquetWriter(f, schema)

        row_group_offsets = np.concatenate([[0], np.cumsum(row_group_sizes)[:-1]])

        for start, size in zip(row_group_offsets, row_group_sizes):
            row_group = pyarrow.record_batch(df.iloc[start : start + size])
            writer.write_batch(row_group)

        writer.close()


def format_experiment_timestamp(timestamp: float):
    time = datetime.fromtimestamp(timestamp)
    year_day = str(int(time.strftime("%j")) + 100)
    day_time = time.strftime("%H%M%S")
    day_second = str(int(day_time[:2]) * 60**2 + int(day_time[2:4]) * 60 + int(day_time[4:])).zfill(5)
    return "".join([year_day, day_second])



from dataclasses import dataclass
from typing import Dict

import numpy as np
from shapely import geometry

from vpt_core import log
from vpt_core.segmentation.geometry_utils import (
    make_polygons_from_label_matrix,
    smooth_and_simplify,
    convert_to_multipoly,
    get_valid_geometry,
)
from vpt_core.segmentation.seg_result import SegmentationResult


@dataclass(frozen=True)
class PolygonCreationParameters:
    simplification_tol: int
    smoothing_radius: int
    minimum_final_area: int


def generate_polygons_from_mask(mask: np.ndarray, polygon_parameters: Dict) -> SegmentationResult:
    log.info("generate_polygons_from_mask")
    parameters = PolygonCreationParameters(**polygon_parameters)
    seg_result = get_polygons_from_mask(mask, parameters.smoothing_radius, parameters.simplification_tol)
    seg_result.remove_polys(lambda poly: poly.area < parameters.minimum_final_area)
    return seg_result


def get_polygons_from_mask(mask: np.ndarray, smoothing_radius, simplification_tolerance) -> SegmentationResult:
    """
    Accepts either a 2D or 3D numpy array label matrix, returns a SegmentationResult of
    MultiPolygons surrounding each label/mask. Performs smoothing and simplification of
    Polygons before returning.
    """

    # If passed 2D data, expand axes so that z-level indexing will work
    if len(mask.shape) == 2:
        mask = np.expand_dims(mask, axis=0)

    polys_data = []
    for z in range(mask.shape[0]):
        list_of_mask_ids = np.unique(mask[z, :, :])
        log.info(f"get_polygons_from_mask: z={z}, labels:{len(list_of_mask_ids)}")
        for idx, mask_id in enumerate(list_of_mask_ids):
            # Value of zero is background, skip
            if mask_id == 0:
                continue

            try:
                # Convert each region of the entity into a polygon
                raw_polys = make_polygons_from_label_matrix(mask_id, mask[z, :, :])

                polys = [
                    smooth_and_simplify(raw_poly, smoothing_radius, simplification_tolerance) for raw_poly in raw_polys
                ]
                polys = [poly for poly in polys if not poly.is_empty]

                # If smoothing and simplifying eliminated all polygons, don't add them to the output
                if len(polys) == 0:
                    continue

                # Transform the list of 1 or more mask polygons into a multipolygon
                multi_poly = convert_to_multipoly(get_valid_geometry(geometry.MultiPolygon(polys)))
                polys_data.append(
                    {
                        SegmentationResult.detection_id_field: idx + 1,
                        SegmentationResult.cell_id_field: mask_id,
                        SegmentationResult.z_index_field: z,
                        SegmentationResult.geometry_field: multi_poly,
                    }
                )
            except ValueError:
                # If the MultiPolygon is not created properly, it is probably because the
                # geometry is low quality or otherwise strange. In that situation, it's ok
                # to discard the geometry by catching the exception and not appending the
                # geometry to the output.
                log.info(f"Mask id {mask_id} could not be converted to a polygon.")

    return SegmentationResult(list_data=polys_data)


def get_upscale_matrix(scale_x: float, scale_y: float) -> np.ndarray:
    return np.array([[scale_x, 0, 0], [0, scale_y, 0], [0, 0, 1]])



import cv2
import numpy as np
from shapely import geometry, make_valid, unary_union

from vpt_core import log


def smooth_and_simplify(poly, radius, tol):
    if isinstance(poly, geometry.MultiPolygon):
        buffered_shapes = (p.buffer(-radius).buffer(radius * 2).buffer(-radius) for p in poly.geoms)
        buffered_multipolygons = (
            p if type(p) is geometry.MultiPolygon else geometry.MultiPolygon([p]) for p in buffered_shapes
        )
        buffered_polygons = (p for mp in buffered_multipolygons for p in mp.geoms)
        poly = geometry.MultiPolygon(buffered_polygons)
    elif isinstance(poly, geometry.Polygon):
        poly = poly.buffer(-radius).buffer(radius * 2).buffer(-radius)
    return largest_geometry(poly.simplify(tolerance=tol))


def largest_geometry(shape):
    """
    If passed a Polygon, returns the Polygon.
    If passed a MultiPolygon, returns the largest Polygon region.
    Else throws TypeError
    """
    if type(shape) is geometry.Polygon:
        return shape
    elif type(shape) is geometry.MultiPolygon:
        if len(shape.geoms) == 0:
            log.warning("Empty multipolygon passed, an empty polygon will be returned")
            return geometry.Polygon()
        sizes = np.array([(idx, g.area) for idx, g in enumerate(shape.geoms)])
        idx_largest = sizes[sizes[:, 1].argmax(), 0]
        return shape.geoms[int(idx_largest)]
    else:
        raise TypeError(f"Objects of type {type(shape)} are not supported")


def convert_to_multipoly(shape):
    if type(shape) is geometry.Polygon:
        return geometry.multipolygon.MultiPolygon([shape])
    elif type(shape) is geometry.MultiPolygon:
        return shape
    elif type(shape) is geometry.GeometryCollection:
        poly_shapes = [g for g in shape.geoms if type(g) in [geometry.Polygon, geometry.MultiPolygon]]
        poly = get_valid_geometry(unary_union(poly_shapes))
        return poly if type(poly) is geometry.MultiPolygon else geometry.multipolygon.MultiPolygon([poly])
    else:
        # If type is not Polygon or Multipolygon, the shape
        # is strange / small and should be rejected
        return geometry.MultiPolygon()


def make_polygons_from_label_matrix(entity_id, label_matrix):
    """
    Creates a list of Polygon objects from a label matrix
    """
    contours, _ = cv2.findContours(
        (label_matrix == entity_id).astype("uint8"),
        cv2.RETR_LIST,
        cv2.CHAIN_APPROX_SIMPLE,
    )
    polys = []
    for c in contours:
        if c.shape[0] < 4:
            continue

        # Create a valid polygon object
        p = geometry.Polygon(c[:, 0, :]).buffer(0)
        p = largest_geometry(p)
        if p.is_empty:
            continue
        polys.append(p)

    return polys


def get_valid_geometry(shape):
    try:
        return make_valid(shape)
    except ValueError:
        return geometry.multipolygon.MultiPolygon()



from abc import ABC, abstractmethod
from typing import Optional, Dict, List, Iterable, Union

import pandas as pd

from vpt_core.io.image import ImageSet
from vpt_core.segmentation.seg_result import SegmentationResult


class SegmentationBase(ABC):
    @staticmethod
    @abstractmethod
    def run_segmentation(
        segmentation_properties: Dict,
        segmentation_parameters: Dict,
        polygon_parameters: Dict,
        result: List[str],
        images: Optional[ImageSet] = None,
        transcripts: Optional[pd.DataFrame] = None,
    ) -> Union[SegmentationResult, Iterable[SegmentationResult]]:
        pass

    @staticmethod
    @abstractmethod
    def validate_task(task: Dict) -> Dict:
        pass



from dataclasses import dataclass
from typing import List

from vpt_core.image import Header


@dataclass
class InputData:
    image_channel: str
    image_preprocessing: List[Header]

    def __init__(self, image_channel: str, image_preprocessing: List[dict]):
        self.image_channel = image_channel
        self.image_preprocessing = [Header(**x) for x in image_preprocessing]



from typing import List, Union, Callable

import cv2
import numpy as np
from skimage import exposure

from vpt_core.image import Filter


def normalization_clahe(image: np.ndarray, clahe_params: dict) -> np.ndarray:
    """
    Normalizes contrast using a CLAHE filter, scales dynamic range, and returns uint8 image
    """
    if image.std() < 0.1:
        return image.astype(np.uint8)
    normalized = exposure.equalize_adapthist(image, **clahe_params)
    subtract = normalized - normalized.min()
    return (
        np.array((subtract / subtract.max()) * 255, dtype=np.uint8)
        if subtract.max() != 0
        else np.zeros(image.shape, dtype=np.uint8)
    )


def normalization_clahe_3d(
    image: Union[np.ndarray, List[np.ndarray]], clahe_params: dict
) -> Union[np.ndarray, List[np.ndarray]]:
    is_3d = clahe_params["image_dimensions"] == "3D"
    cur_args = dict(clahe_params)
    cur_args.pop("image_dimensions")
    if isinstance(image, list):
        if is_3d:
            if len(cur_args["kernel_size"]) == 2:
                cur_args["kernel_size"] = [3, *cur_args["kernel_size"]]
            res = normalization_clahe(np.array(image), cur_args)
            return [res_im for res_im in res]
        else:
            return [normalization_clahe(im, cur_args) for im in image]
    return normalization_clahe(image, cur_args)


def filter_iterative(
    image: Union[np.ndarray, List[np.ndarray]], filter_callback: Callable, *args, **kwargs
) -> Union[np.ndarray, List[np.ndarray]]:
    if isinstance(image, list):
        return [filter_callback(im, *args, **kwargs) for im in image]
    return filter_callback(image, *args, **kwargs)


def normalize(image: np.ndarray) -> np.ndarray:
    return cv2.normalize(image, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)


def create_normalization_filter(p: dict) -> Filter:
    norm_type = p.get("type", "default")
    if norm_type == "CLAHE":
        args = {
            "clip_limit": p.get("clip_limit", 0.01),
            "kernel_size": np.array(p.get("filter_size", [100, 100])),
            "image_dimensions": p.get("image_dimensions", "2D"),
        }
        if args["kernel_size"].ndim != 1:
            raise ValueError("Invalid filter parameter: kernel size should be a one-dimensional array")
        if len(args["kernel_size"]) < 2:
            raise ValueError("Invalid filter parameter: kernel size should consist of at least two elements")
        return lambda img: normalization_clahe_3d(img, args)
    elif norm_type == "default":
        return lambda img: filter_iterative(img, normalize)
    raise TypeError(f"unsupported normalization type {norm_type}")


def create_blur_filter(p: dict) -> Filter:
    sz = p.get("size", 5)
    blur_type = p.get("type", "average")
    if blur_type == "median":
        return lambda image: filter_iterative(image, cv2.medianBlur, ksize=sz)
    elif blur_type == "gaussian":
        return lambda image: filter_iterative(image, cv2.GaussianBlur, (sz, sz), 0)
    elif blur_type == "average":
        return lambda image: filter_iterative(image, cv2.blur, (sz, sz))
    raise TypeError(f"unsupported blur type {blur_type}")


def create_downsample_filter(p: dict) -> Filter:
    scale = p.get("scale", 2)
    return lambda image: filter_iterative(image, cv2.resize, (0, 0), fx=1 / scale, fy=1 / scale)


def _filter_from_merlin(s_image: np.ndarray, thr: int) -> np.ndarray:
    ret = s_image
    if s_image.dtype == np.uint16:
        ret = cv2.convertScaleAbs(s_image, alpha=(255.0 / 65535.0))
    ret = cv2.adaptiveThreshold(ret, 1, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, thr, -1)
    ret = ret * s_image
    return ret


def create_merlin_watershed_filter(p: dict) -> Filter:
    threshold = int(255 * p.get("threshold", 0.5))
    return lambda image: filter_iterative(image, _filter_from_merlin, threshold)


def apply_sequence(
    image: Union[np.ndarray, List[np.ndarray]], filters: List[Filter]
) -> Union[np.ndarray, List[np.ndarray]]:
    for f in filters:
        if f is not None:
            image = f(image)
    return image



from __future__ import annotations

from collections import defaultdict, namedtuple
from typing import List, Optional, Set, Tuple, Callable, Iterable

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from shapely import affinity, geometry, MultiPolygon
from shapely.ops import unary_union

from vpt_core import log
from vpt_core.segmentation import ResultFields
from vpt_core.segmentation.geometry_utils import get_valid_geometry, convert_to_multipoly
from vpt_core.segmentation.segmentation_item import SegmentationItem


class SegmentationResult(ResultFields):
    MAX_ENTITY_ID = 99999
    MAX_TILE_ID = 99999
    MAX_TASK_ID = 9

    def __init__(
        self,
        list_data=None,
        dataframe: Optional[GeoDataFrame] = None,
        entity: Optional[str] = None,
    ):
        if list_data is None and dataframe is None:
            list_data = []
        if dataframe is not None:
            self.df = dataframe.copy()
        else:
            self.df = GeoDataFrame(
                list_data,
                columns=[
                    self.detection_id_field,
                    self.cell_id_field,
                    self.z_index_field,
                    self.geometry_field,
                ],
            )
        if entity is not None:
            self.entity_type = entity
        self.df.set_geometry(self.geometry_field, inplace=True)

        if self.parent_entity_field not in self.df.columns:
            self.set_column(self.parent_entity_field, None)
        if self.parent_id_field not in self.df.columns:
            self.set_column(self.parent_id_field, None)

        self.df.loc[:, self.cell_id_field] = self.df.loc[:, self.cell_id_field].astype("int64")

    def item(self, item_id: int) -> SegmentationItem:
        df = self.df.loc[self.df[self.cell_id_field] == item_id]
        if len(df) < 1:
            raise KeyError(f"invalid item id {item_id}")

        return SegmentationItem(self.entity_type if hasattr(self, "entity_type") else "", df)

    @staticmethod
    def get_connected_items(processed_item: SegmentationItem) -> List[SegmentationItem]:
        data = SegmentationResult._separate_multi_geometries(processed_item.df)
        seg_res = SegmentationResult(dataframe=data)
        seg_res.fuse_across_z(coverage_threshold=-0.1)
        items = []
        for cell_id in seg_res.df[processed_item.cell_id_field].unique():
            item = seg_res.item(cell_id)
            item.set_entity_type(processed_item.get_entity_type())
            item.df = item.df.assign(**{processed_item.cell_id_field: processed_item.get_entity_id()})
            items.append(item)
        return items

    def update_item(self, item: SegmentationItem):
        clone = item.as_copy()
        self.df = pd.concat(
            [self.df.drop(self.df.loc[self.df[self.cell_id_field] == item.get_entity_id()].index), clone.df],
            ignore_index=True,
        )

    def add_item(self, item: SegmentationItem) -> int:
        clone = item.as_copy()
        clone.set_entity_type(self.entity_type)
        # todo implement new id generation
        if len(self.df) > 0:
            new_id = self.df[SegmentationItem.cell_id_field].max() + 1
        else:
            entity_id_len = len(str(SegmentationResult.MAX_ENTITY_ID))
            id_prefix = str(item.get_entity_id())[:-entity_id_len]
            new_id = np.int64(id_prefix + "0".zfill(entity_id_len))
        clone.df[SegmentationItem.cell_id_field] = clone.df[SegmentationItem.cell_id_field].apply(lambda x: new_id)
        self.df = pd.concat([self.df, clone.df], ignore_index=True)
        return new_id

    def update_column(self, column_name: str, callback: Callable, *args, **kwargs):
        self.df[column_name] = self.df[column_name].apply(lambda x: callback(x, *args, **kwargs), convert_dtype=False)

    def update_geometry(self, callback: Callable, *args, **kwargs):
        def transform(x) -> geometry.MultiPolygon:
            return convert_to_multipoly(callback(x, *args, **kwargs))

        self.df[self.geometry_field] = self.df[self.geometry_field].apply(transform)

    def set_column(self, column_name: str, data):
        kwargs = {column_name: data}
        self.df = self.df.assign(**kwargs)

    def remove_column(self, column_name):
        if column_name in self.df.columns:
            self.df.drop(columns=[column_name], inplace=True)

    def set_entity_type(self, entity: str):
        self.entity_type = entity

    def translate_geoms(self, dx, dy):
        self.update_geometry(affinity.translate, xoff=dx, yoff=dy)

    def transform_geoms(self, matrix):
        tform_flat = [*matrix[:2, :2].flatten(), *matrix[:2, 2].flatten()]
        self.update_geometry(affinity.affine_transform, matrix=tform_flat)

    def remove_polys(self, condition, *args, **kwargs):
        indices_to_remove = []
        for idx, row in self.df.iterrows():
            if condition(row[self.geometry_field], *args, **kwargs):
                indices_to_remove.append(idx)
        if len(indices_to_remove) > 0:
            self.df.drop(indices_to_remove, inplace=True)

    def _get_replication_pairs(self, z_planes: Set[int]):
        copy_pairs: List[Tuple[int, int]] = []
        if len(self.df) < 1:
            return copy_pairs
        occupied = set(self.df["ZIndex"].to_list())
        exceptional_z = min(z_planes) - 1
        while z_planes != occupied:
            missing_layers = z_planes.difference(occupied)
            neighbor_layers = [
                [present if abs(present - missing) == 1 else exceptional_z for present in occupied]
                for missing in missing_layers
            ]
            layer_pairs = [
                (missing, max(neighbor))
                for missing, neighbor in zip(missing_layers, neighbor_layers)
                if any([z != exceptional_z for z in neighbor])
            ]
            if len(layer_pairs) == 0:
                return copy_pairs

            copy_pairs.extend(layer_pairs)
            for x in layer_pairs:
                occupied.add(x[0])
        return copy_pairs

    def replicate_across_z(self, z_planes: List[int]):
        replicate_pairs = self._get_replication_pairs(set(z_planes))
        for missing_z, copy_z in replicate_pairs:
            missing_z_seg = SegmentationResult(dataframe=self.df.loc[self.df[self.z_index_field] == copy_z])
            missing_z_seg.set_column(self.z_index_field, missing_z)
            self.df = self.combine_segmentations([self, missing_z_seg]).df

    def set_z_levels(self, z_levels: List[float], column_name):
        levels = [z_levels[row[self.z_index_field]] for _, row in self.df.iterrows()]
        self.set_column(column_name, levels)

    def _assign_entity_index(
        self, entities_df: GeoDataFrame, entities_ids: List, cur_poly: MultiPolygon, coverage_threshold: float
    ) -> List[int]:
        res = []
        geometry_col = self.df.columns.get_loc(self.geometry_field)
        for i in entities_ids:
            prev_poly = entities_df[entities_df[self.cell_id_field] == i].iloc[0, geometry_col]
            polys_intersection = prev_poly.intersection(cur_poly)
            if (polys_intersection.area / min(prev_poly.area, cur_poly.area)) > coverage_threshold:
                res.append(i)
        return res

    def _update_entities_index(self, overlaps_pairs: np.ndarray, cur_cells_ids: List, updated_ids: List, z_index: int):
        cur_df = self.df.loc[self.df[self.z_index_field] == z_index]
        for old_id, new_id in updated_ids:
            for j in cur_df.loc[cur_df[self.cell_id_field] == old_id].index:
                self.df.at[j, self.cell_id_field] = new_id
                cur_df.at[j, self.cell_id_field] = new_id
            for i in [index for index in range(len(cur_cells_ids)) if cur_cells_ids[index] == old_id]:
                cur_cells_ids[i] = new_id

            if len(overlaps_pairs) == 0:
                return
            overlaps_pairs[overlaps_pairs[:, 0] == old_id, 0] = new_id

    def fuse_across_z(self, coverage_threshold: float = 0.5):
        z_lines = self.df["ZIndex"].unique()
        reserved_ids = []
        geometry_col = self.df.columns.get_loc(self.geometry_field)
        for z_i in range(1, len(z_lines)):
            prev_df = self.df.loc[self.df[self.z_index_field] == z_lines[z_i - 1]]
            prev_df = prev_df.assign(**{self.z_index_field: z_lines[z_i]})
            cur_df = self.df.loc[self.df[self.z_index_field] == z_lines[z_i]]

            reserved_ids.extend(prev_df[self.cell_id_field].to_list())
            cur_cells_ids = cur_df[self.cell_id_field].unique()
            overlaps = np.array(self.find_overlapping_entities(prev_df, cur_df))

            for i in range(len(cur_cells_ids)):
                cur_i = cur_cells_ids[i]
                cur_intersected_i_to_update = []
                need_new_index = True
                if len(overlaps) > 0 and cur_i in overlaps[:, 0]:
                    cur_poly = cur_df[cur_df[self.cell_id_field] == cur_i].iloc[0, geometry_col]
                    prev_ids = [pair[1] for pair in overlaps if pair[0] == cur_i]

                    prev_ids = self._assign_entity_index(prev_df, prev_ids, cur_poly, coverage_threshold)
                    if len(prev_ids) > 1:
                        prev_i_to_update = [(prev_i, prev_ids[0]) for prev_i in prev_ids[1:]]
                        for z in z_lines[:z_i]:
                            self._update_entities_index(np.array([]), [], prev_i_to_update, z)

                    if len(prev_ids) > 0:
                        prev_i = prev_ids[0]
                        need_new_index = False
                        if prev_i != cur_i:
                            if prev_i in cur_cells_ids:
                                new_id = max(max(reserved_ids), max(cur_cells_ids)) + 1
                                cur_intersected_i_to_update.append((prev_i, new_id))
                            cur_intersected_i_to_update.append((cur_i, prev_i))

                if need_new_index:
                    if cur_i in reserved_ids:
                        new_id = max(max(reserved_ids), max(cur_cells_ids)) + 1
                        cur_intersected_i_to_update.append((cur_i, new_id))
                self._update_entities_index(overlaps, cur_cells_ids, cur_intersected_i_to_update, z_lines[z_i])
                cur_df = self.df.loc[self.df[self.z_index_field] == z_lines[z_i]]

        self.group_duplicated_entities()

    def group_duplicated_entities(self):
        for z, df in self.df.groupby(self.z_index_field):
            for entity_id, entity_df in df.groupby(self.cell_id_field):
                if len(entity_df) == 1:
                    continue
                poly = convert_to_multipoly(get_valid_geometry(unary_union(entity_df[self.geometry_field])))
                self.df.at[entity_df.index[0], self.geometry_field] = poly
                self.df.drop(entity_df.index[1:], inplace=True)

    @staticmethod
    def _separate_multi_geometries(data: GeoDataFrame) -> GeoDataFrame:
        geoms_separated = data.copy()
        geometry_field = SegmentationResult.geometry_field
        new_rows = []
        for index, row in geoms_separated.iterrows():
            i = 0
            for geom in row[geometry_field].geoms:
                if i == 0:
                    i += 1
                    geoms_separated.at[index, geometry_field] = convert_to_multipoly(get_valid_geometry(geom))
                    continue
                new_row = geoms_separated[geoms_separated.index == index].copy()
                new_row[geometry_field] = convert_to_multipoly(get_valid_geometry(geom))
                new_rows.append(new_row)
        geoms_separated = pd.concat([geoms_separated, *new_rows], ignore_index=True)
        geoms_separated = geoms_separated.assign(**{SegmentationResult.cell_id_field: range(len(geoms_separated))})
        return geoms_separated

    def remove_edge_polys(self, tile_size: Tuple[int, int]):
        crop_frame = geometry.box(-10, -10, tile_size[0] + 10, tile_size[1] + 10) - geometry.box(
            5, 5, tile_size[0] - 5, tile_size[1] - 5
        )
        for entity_id in self.df[self.cell_id_field].unique():
            cur_gdf = self.df.loc[self.df[self.cell_id_field] == entity_id]
            is_intersected = False
            for z in cur_gdf[self.z_index_field]:
                cur_geom_df = cur_gdf.loc[cur_gdf[self.z_index_field] == z, self.geometry_field]
                if len(cur_geom_df) == 0:
                    continue
                if any(cur_geom_df.values.intersects(crop_frame)):
                    is_intersected = True
                    break
            if is_intersected:
                self.df = self.df.drop(self.df.loc[self.df[self.cell_id_field] == entity_id].index)

    def _resolve_cell_overlap(self, large_cell: GeoDataFrame, small_cell: GeoDataFrame, min_distance: int):
        """
        Trims area from larger entity
        """
        z_planes_in_both = set(large_cell[self.z_index_field]).intersection(set(small_cell[self.z_index_field]))

        for z in z_planes_in_both:
            # Gets the large and small geometries
            large_at_z_idx = large_cell.loc[large_cell[self.z_index_field] == z].index[0]
            small_at_z_idx = small_cell.loc[small_cell[self.z_index_field] == z].index[0]

            large_at_z = self.df.at[large_at_z_idx, self.geometry_field]
            small_at_z = self.df.at[small_at_z_idx, self.geometry_field]

            try:
                # Trims the larger geometry with a small buffer
                valid_large = get_valid_geometry(large_at_z)
                valid_small = get_valid_geometry(small_at_z)

                trimmed_raw = valid_large.difference(valid_small.buffer(min_distance))
                trimmed_geometry = convert_to_multipoly(get_valid_geometry(trimmed_raw))
            except ValueError:
                log.info("Entity could not be converted to a valid polygon, removing.")
                trimmed_geometry = geometry.MultiPolygon()

            # Overwrites large geometry with trimmed geometry
            self.df.at[large_at_z_idx, self.geometry_field] = trimmed_geometry

    @staticmethod
    def find_overlapping_entities(dataframe_1: GeoDataFrame, dataframe_2: Optional[GeoDataFrame] = None) -> List:
        """
        Uses shapely library to rapidly identify entities that overlap in 3D
        """

        gdf2 = dataframe_2 if dataframe_2 is not None else dataframe_1

        overlaps = dataframe_1.sindex.query_bulk(gdf2.geometry, predicate="intersects").T

        # Remove self-intersections
        if dataframe_2 is None:
            overlaps = np.array([pair for pair in overlaps if pair[0] != pair[1]])
        if len(overlaps) == 0:
            return []

        z_field = SegmentationResult.z_index_field
        id_field = SegmentationResult.cell_id_field

        same_z_overlaps = gdf2[z_field].values[overlaps[:, 0]] == dataframe_1[z_field].values[overlaps[:, 1]]

        # Map list positions back to Entity IDs
        entity_overlap = np.transpose(
            [gdf2[id_field].values[overlaps[:, 0]], dataframe_1[id_field].values[overlaps[:, 1]]]
        )[same_z_overlaps]

        res: List = []
        # Remove repeated sets of EntityIDs
        if dataframe_2 is None:
            res = list(set(frozenset(pair) for pair in entity_overlap))
        else:
            res = list(set(tuple(pair) for pair in entity_overlap))
        return res

    def get_volume(self):
        return np.sum([x.area for x in self.df[self.geometry_field]])

    def get_z_planes(self) -> Set:
        return set(self.df[self.z_index_field])

    def get_z_geoms(self, z_line: int) -> GeoSeries:
        return self.df.loc[self.df[self.z_index_field] == z_line, self.geometry_field]

    @staticmethod
    def combine_segmentations(segmentations: List):
        non_empty_segmentations = [seg for seg in segmentations if len(seg.df) > 0]
        if len(non_empty_segmentations) > 1:
            to_concat = [seg.df for seg in non_empty_segmentations]
            df = GeoDataFrame(pd.concat(to_concat, ignore_index=True))
            duplicated_fields = [
                SegmentationResult.cell_id_field,
                SegmentationResult.z_index_field,
            ]
            deprecated_indexes = df[df.duplicated(duplicated_fields)].index
            if len(deprecated_indexes) > 0:
                log.warning(
                    f"Found several entity rows for the same z planes. {len(deprecated_indexes)} rows will be "
                    f"removed from segmentation result."
                )
                df.drop(deprecated_indexes, inplace=True)
            res = SegmentationResult(dataframe=df)
            res.set_column(SegmentationResult.detection_id_field, list(range(len(res.df))))
            return res
        else:
            return non_empty_segmentations[0] if len(non_empty_segmentations) > 0 else SegmentationResult()

    def _union_entities(self, base_gdf, add_gdf):
        """
        Adds the area from entity_id_add to entity_id_base across z-levels
        """

        # Area of intersection
        add_entity = dict(zip(add_gdf[self.z_index_field], add_gdf[self.geometry_field]))

        occupied_z_planes = set(base_gdf[self.z_index_field]).intersection(set(add_entity))
        for z in occupied_z_planes:
            base_geoms = base_gdf.loc[base_gdf[self.z_index_field] == z]
            if len(base_geoms) == 0:
                continue
            if z not in add_entity:
                continue
            base_geom_z_idx = base_geoms.index[0]

            base_geom = base_gdf.at[base_geom_z_idx, self.geometry_field]
            add_geom = add_entity[z]

            try:
                valid_base = get_valid_geometry(base_geom)
                valid_add = get_valid_geometry(add_geom)

                union_raw = valid_base.union(valid_add)
                union_geom = convert_to_multipoly(get_valid_geometry(union_raw))
            except ValueError:
                log.info("Entity could not be converted to a valid polygon, removing.")
                union_geom = geometry.MultiPolygon()

            # Overwrites large geometry with trimmed geometry
            self.df.at[base_geom_z_idx, self.geometry_field] = union_geom

    def cell_size_filter(self, minimum_area: int):
        depricated_entity_ids = []
        for entity_id, gdf in self.df.groupby(self.cell_id_field):
            if len(gdf) > 0 and not any(cell.area > minimum_area for cell in gdf[self.geometry_field]):
                depricated_entity_ids.append(entity_id)
        self._drop_by_entity_id(depricated_entity_ids)

    def make_non_overlapping_polys(self, min_distance: int = 2, min_area: int = 100, log_progress: bool = False):
        # Find cells that have any overlapping area
        problem_sets = self.find_overlapping_entities(self.df)
        log.info(f"Found {len(problem_sets)} overlaps")

        # For each pair of overlapping cells, resolve the conflict
        depricated_entity_ids = []

        def iterate(x: Iterable):
            if log_progress:
                return log.show_progress(x)
            else:
                return x

        # Union the large overlap Entities
        for problem in iterate(problem_sets):
            # Get the slice of the dataframe for each entity
            pl = list(problem)
            entity_id_left, entity_id_right = pl[0], pl[1]

            # If either cell is in the depricated id list, ignore the overlap.
            # The entity union / drop process may drop a cell that should be retained.
            if entity_id_left in depricated_entity_ids or entity_id_right in depricated_entity_ids:
                continue
            left = self.item(entity_id_left)
            right = self.item(entity_id_right)

            # Find seperate and overlapping volumes of the cells
            volume_left = left.get_volume()
            volume_right = right.get_volume()
            overlap_volume = left.get_overlapping_volume(right)
            overlap_volume_percent = overlap_volume / min(volume_right, volume_left)

            # If overlap is > 50% of either cell, eliminate the small cell and keep the big one
            if overlap_volume_percent > 0.5:
                if volume_left > volume_right:
                    self._union_entities(left.df, right.df)
                    depricated_entity_ids.append(entity_id_right)
                else:
                    self._union_entities(right.df, left.df)
                    depricated_entity_ids.append(entity_id_left)

        # With large overlaps resolved, re-identify problem sets and trim overlaps
        self._drop_by_entity_id(depricated_entity_ids)
        problem_sets = self.find_overlapping_entities(self.df)
        log.info(f"After union of large overlaps, found {len(problem_sets)} overlaps")

        for problem in iterate(problem_sets):
            # Get the slice of the dataframe for each entity
            pl = list(problem)
            entity_id_left, entity_id_right = pl[0], pl[1]
            left = self.item(entity_id_left)
            right = self.item(entity_id_right)

            # Find seperate and overlapping volumes of the cells
            volume_left = left.get_volume()
            volume_right = right.get_volume()

            # Trim the larger cell to dodge the smaller cell
            if volume_left > volume_right:
                self._resolve_cell_overlap(left.df, right.df, min_distance)
            else:
                self._resolve_cell_overlap(right.df, left.df, min_distance)

        # After both steps, check for any remaining overlaps
        problem_sets = self.find_overlapping_entities(self.df)
        log.info(f"After both resolution steps, found {len(problem_sets)} uncaught overlaps")

        # Filter any small cells that were created
        self.cell_size_filter(min_area)

    def union_intersections(self, min_distance: int, min_area: int):
        problem_sets = self.find_overlapping_entities(self.df)
        depricated_entity_ids = []
        for problem in problem_sets:
            entity_id_left, entity_id_right, *_ = list(problem)
            left_df = self.df.loc[self.df[self.cell_id_field] == entity_id_left]
            right_df = self.df.loc[self.df[self.cell_id_field] == entity_id_right]
            self._union_entities(left_df, right_df)
            depricated_entity_ids.append(entity_id_right)
        self._drop_by_entity_id(depricated_entity_ids)
        self.cell_size_filter(min_area)

    def larger_resolve_intersections(self, min_distance: int, min_area: int):
        problem_sets = self.find_overlapping_entities(self.df)
        depricated_entity_ids = []
        for problem in problem_sets:
            entity_id_left, entity_id_right, *_ = list(problem)
            left = self.item(entity_id_left)
            right = self.item(entity_id_right)
            left_area, right_area = left.get_volume(), right.get_volume()
            if left_area > right_area:
                depricated_entity_ids.append(entity_id_right)
            else:
                depricated_entity_ids.append(entity_id_left)
        self._drop_by_entity_id(depricated_entity_ids)
        self.cell_size_filter(min_area)

    def _drop_by_entity_id(self, entity_ids_to_delete):
        depricated_row_indexes = []
        for entity_id in entity_ids_to_delete:
            row_indexes = self.df.loc[self.df[self.cell_id_field] == entity_id].index.to_list()
            depricated_row_indexes.extend(row_indexes)
        self.df.drop(depricated_row_indexes, inplace=True)

    @staticmethod
    def reindex_by_task(tasks_results: List[SegmentationResult], tasks_numbers: List[int]):
        def add_task_to_entity_id(old_id, task_number) -> np.int64:
            entity_str_len = len(str(SegmentationResult.MAX_ENTITY_ID))
            task_str_len = len(str(SegmentationResult.MAX_TASK_ID))

            old_id_str = str(old_id).zfill(entity_str_len)
            task_str = str(task_number + 1).zfill(task_str_len)

            if len(old_id_str) > entity_str_len or len(task_str) > task_str_len:
                raise OverflowError("EntityID is out of the int64 type range")
            return np.int64("".join([task_str, old_id_str]))

        for task_i in range(len(tasks_results)):
            tasks_results[task_i].update_column(
                SegmentationResult.cell_id_field,
                add_task_to_entity_id,
                task_number=tasks_numbers[task_i],
            )
        return tasks_results

    def create_relationships(self, parent_seg: SegmentationResult, coverage_threshold: float):
        if len(parent_seg.df) == 0 or len(self.df) == 0:
            return
        problem_sets = self.find_overlapping_entities(self.df, parent_seg.df)

        # collecting possible relations for each child entity
        relationships = defaultdict(list)
        RelationInfo = namedtuple("RelationInfo", "parent_id child_rate parent_rate overlap")
        for problem in problem_sets:
            pl = list(problem)
            parent_id, child_id = pl[0], pl[1]

            child = self.item(child_id)
            parent_entity = parent_seg.item(parent_id)

            overlapping = parent_entity.get_overlapping_volume(child)
            parent_rate = overlapping / parent_entity.get_volume()
            child_rate = overlapping / child.get_volume()
            if max(child_rate, parent_rate) > coverage_threshold:
                relationships[child_id].append(RelationInfo(parent_id, child_rate, parent_rate, overlapping))

        # get from the possible relations the child entities that are smaller than the parent entities
        result = {child_id: max(value, key=lambda x: x.child_rate) for child_id, value in relationships.items()}
        result = {
            child_id: value.parent_id for child_id, value in result.items() if value.child_rate > coverage_threshold
        }

        # get from the possible relations the child entities that are larger than the parent entities
        other = {
            child_id: max(value, key=lambda x: x.overlap)
            for child_id, value in relationships.items()
            if child_id not in result
        }
        result.update(
            {
                child_id: value.parent_id
                for child_id, value in other.items()
                if value.parent_id not in result.values() and value.parent_rate > coverage_threshold
            }
        )

        parent_ids = pd.array(
            [result.get(row[self.cell_id_field]) for _, row in self.df.iterrows()],
            dtype="object",
        )
        parent_types = [parent_seg.entity_type if parent_id is not None else None for parent_id in parent_ids]
        self.set_column(self.parent_id_field, data=parent_ids)
        self.set_column(self.parent_entity_field, data=parent_types)



import geopandas as gpd
from geopandas.geodataframe import DEFAULT_GEO_COLUMN_NAME
from pyarrow import parquet, Table
from shapely import wkb

from vpt_core.io.vzgfs import vzg_open, retrying_attempts
from vpt_core.segmentation.seg_result import SegmentationResult


def pyarrow_table_to_pandas(table: Table):
    df = gpd.GeoDataFrame(table.to_pandas(integer_object_nulls=True))
    if SegmentationResult.geometry_field in df.columns:
        df[SegmentationResult.geometry_field] = df[SegmentationResult.geometry_field].apply(wkb.loads)
        df.set_geometry(SegmentationResult.geometry_field, inplace=True)
    else:
        df[DEFAULT_GEO_COLUMN_NAME] = df[DEFAULT_GEO_COLUMN_NAME].apply(wkb.loads)
    if SegmentationResult.parent_id_field in df.columns:
        df.loc[:, SegmentationResult.parent_id_field] = df.loc[:, SegmentationResult.parent_id_field].astype("object")
    return df


def read_parquet(path: str) -> gpd.GeoDataFrame:
    for attempt in retrying_attempts():
        with attempt, vzg_open(path, "rb") as f:
            pq = parquet.ParquetFile(f)
            return pyarrow_table_to_pandas(pq.read())



from typing import Callable, Dict, List

from vpt_core.image import Filter, Header
from vpt_core.image.filter import (
    apply_sequence,
    create_blur_filter,
    create_downsample_filter,
    create_merlin_watershed_filter,
    create_normalization_filter,
)

factory_map: Dict[str, Callable[[dict], Filter]] = {
    "normalize": create_normalization_filter,
    "blur": create_blur_filter,
    "downsample": create_downsample_filter,
    "merlin-ws": create_merlin_watershed_filter,
}


def create_filter(h: Header) -> Filter:
    if h.name in factory_map:
        return factory_map[h.name](h.parameters)
    else:
        raise NameError(f"invalid filter name {h.name}")


def create_filter_by_sequence(headers: List[Header]) -> Filter:
    filters = [create_filter(h) for h in headers] if headers else []
    return lambda image: apply_sequence(image, filters)



from typing import List, Optional

import numpy as np
import pandas as pd
from shapely import MultiPolygon, box

from vpt_core.segmentation.seg_result import SegmentationResult


def Square(x, y, a) -> MultiPolygon:
    return MultiPolygon([box(x, y, x + a, y + a)])


def Rect(x, y, a, b) -> MultiPolygon:
    return MultiPolygon([box(x, y, x + a, y + b)])


def from_shapes(
    geoms: List[MultiPolygon], cids: Optional[List[int]] = None, entity_type: str = "cell"
) -> SegmentationResult:
    ids = range(len(geoms))
    res = SegmentationResult(
        list_data={
            SegmentationResult.detection_id_field: list(ids),
            SegmentationResult.z_index_field: [0 for _ in ids],
            SegmentationResult.cell_id_field: cids if cids else list(ids),
            SegmentationResult.geometry_field: geoms,
        },
        entity=entity_type,
    )
    res.set_column(res.entity_name_field, res.entity_type)
    return res


def from_shapes_3d(
    geoms: List[List[MultiPolygon]], cids: Optional[List[int]] = None, entity_type: str = "cell"
) -> SegmentationResult:
    zs = [z for cell in geoms for z in range(len(cell))]
    ci = cids if cids else [cell_id for cell_id, cell in enumerate(geoms) for _ in cell]
    res = SegmentationResult(
        list_data={
            SegmentationResult.detection_id_field: list(range(sum(len(x) for x in geoms))),
            SegmentationResult.z_index_field: zs,
            SegmentationResult.cell_id_field: ci,
            SegmentationResult.geometry_field: [g for cell in geoms for g in cell],
        },
        entity=entity_type,
    )
    res.set_column(res.entity_name_field, res.entity_type)
    return res


def assert_df_equals(gdf1, gdf2, area_epsilon: float = 100):
    gdf1_sorted = gdf1.sort_values([SegmentationResult.cell_id_field, SegmentationResult.z_index_field])
    gdf2_sorted = gdf2.sort_values([SegmentationResult.cell_id_field, SegmentationResult.z_index_field])

    iloc1 = gdf1_sorted.drop(
        [SegmentationResult.geometry_field, SegmentationResult.detection_id_field],
        axis=1,
    ).reset_index(drop=True)
    iloc2 = gdf2_sorted.drop(
        [SegmentationResult.geometry_field, SegmentationResult.detection_id_field],
        axis=1,
    ).reset_index(drop=True)

    assert np.logical_or(iloc1 == iloc2, np.logical_and(pd.isnull(iloc1), pd.isnull(iloc2))).all().all()
    v1 = gdf1_sorted[SegmentationResult.geometry_field].values
    v2 = gdf2_sorted[SegmentationResult.geometry_field].values
    assert (v1.symmetric_difference(v2).area < area_epsilon).all()


def assert_seg_equals(s1: SegmentationResult, s2: SegmentationResult, area_epsilon: float = 100):
    assert_df_equals(s1.df, s2.df, area_epsilon)
